import xml.etree.ElementTree as ET
import json
import hashlib
from multiprocessing import Process, resource_tracker
from multiprocessing import shared_memory


def remove_shm_from_resource_tracker():
    """Monkey-patch multiprocessing.resource_tracker so SharedMemory won't be tracked

    More details at: https://bugs.python.org/issue38119
    """

    def fix_register(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.register(self, name, rtype)
    resource_tracker.register = fix_register

    def fix_unregister(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.unregister(self, name, rtype)
    resource_tracker.unregister = fix_unregister

    if "shared_memory" in resource_tracker._CLEANUP_FUNCS:
        del resource_tracker._CLEANUP_FUNCS["shared_memory"]


def write_to_data_structure(self, tag_name, tag_data):
    tag_start_pos = (self.data_dict["tag_dict"][tag_name]["START_POS"]) + self.dict_end_pos
    tag_end_pos = (self.data_dict["tag_dict"][tag_name]["END_POS"]) + self.dict_end_pos + 1
    tag_len = (self.data_dict["tag_dict"][tag_name]["LENGTH"])
    tag_data_str = str(tag_data)

    if len(tag_data_str) <= tag_len:
        encoded_tag = bytes(f"{tag_data_str:<{tag_len}}".encode('utf-8'))
        self.shm.buf[tag_start_pos:tag_end_pos] = encoded_tag

    else:
        raise ValueError("Tag length exceed structure data size")


def read_data_structure(self):
    if self.data_type_file:
        tree = self.data_type_file

    else:
        test_path = self.data_type.split(".")
        if len(test_path) == 1:
            data_type_path = "data_structures/" + self.data_type + ".xml"
        else:
            data_type_path = self.data_type
        tree = ET.parse(data_type_path)

    data_structure_dict = {}
    combined_pos = 0

    root = tree.getroot()
    data_structure_dict["tag_dict"] = {}
    for child in root:
        data_structure_dict["tag_dict"][child.tag] = {}
        data_structure_dict["tag_dict"][child.tag]["LENGTH"] = int(child.attrib["LENGTH"])
        data_structure_dict["tag_dict"][child.tag]["START_POS"] = combined_pos
        data_structure_dict["tag_dict"][child.tag]["END_POS"] = combined_pos + (int(child.attrib["LENGTH"]) - 1)
        if 'INIT_VALUE' in child.attrib:
            data_structure_dict["tag_dict"][child.tag]['INIT_VALUE'] = child.attrib['INIT_VALUE']
        else:
            data_structure_dict["tag_dict"][child.tag]['INIT_VALUE'] = None

        if 'TYPE' in child.attrib:
            data_structure_dict["tag_dict"][child.tag]['TYPE'] = child.attrib['TYPE']
        else:
            data_structure_dict["tag_dict"][child.tag]['TYPE'] = None

        combined_pos = combined_pos + (int(child.attrib["LENGTH"]))
    data_structure_dict["tag_len"] = combined_pos
    return data_structure_dict


class RamShare:
    def __init__(self, share_name, data_type=None, data_type_file=None):
        self.share_name = share_name
        if data_type or data_type_file:
            self.data_type = data_type
            self.data_type_file = data_type_file
            self.data_dict = read_data_structure(self)
            self.data_dict_bytes = json.dumps(self.data_dict).encode('utf-8')
            self.data_dict_checksum = hashlib.md5(self.data_dict_bytes).hexdigest()
            data_dict_checksum_bytes = json.dumps(self.data_dict_checksum).encode('utf-8')
            self.dict_end_pos = len(self.data_dict_bytes) + 45
            shared_size = self.dict_end_pos + self.data_dict["tag_len"]
            encoded_end_dict_pos = f"{str(self.dict_end_pos):<{9}}".encode('utf-8')

            try:
                remove_shm_from_resource_tracker()
                self.shm = shared_memory.SharedMemory(create=True, size=shared_size, name=self.share_name)
                self.shm.buf[0:34] = data_dict_checksum_bytes
                self.shm.buf[35:44] = bytes(encoded_end_dict_pos)
                self.shm.buf[45:len(self.data_dict_bytes) + 45] = self.data_dict_bytes

                for key in self.data_dict["tag_dict"]:
                    if self.data_dict["tag_dict"][key]["INIT_VALUE"]:
                        write_to_data_structure(self, key, self.data_dict["tag_dict"][key]["INIT_VALUE"])


            except:
                remove_shm_from_resource_tracker()
                self.shm = shared_memory.SharedMemory(name=self.share_name)
                shared_checksum_bytes = bytes(self.shm.buf[0:34])
                shared_checksum = json.loads(shared_checksum_bytes.decode('utf-8'))
                if self.data_dict_checksum != shared_checksum:
                    raise ValueError("Data Structure Mismatch")
                pass

        else:
            try:
                remove_shm_from_resource_tracker()
                self.shm = shared_memory.SharedMemory(name=self.share_name)
                self.dict_end_pos = bytes(self.shm.buf[35:44])
                self.dict_end_pos = int(self.dict_end_pos.decode("utf-8"))
                self.dict_bytes = bytes(self.shm.buf[45:self.dict_end_pos])
                self.data_dict = json.loads(self.dict_bytes.decode('utf-8'))

            except:
                raise ValueError("Data Structure missing")
                pass

    def read_tag(self, tag_name):
        tag_start_pos = (self.data_dict["tag_dict"][tag_name]["START_POS"]) + self.dict_end_pos
        tag_end_pos = (self.data_dict["tag_dict"][tag_name]["END_POS"]) + self.dict_end_pos
        tag_output_bytes = bytes(self.shm.buf[tag_start_pos:tag_end_pos])
        tag_output = tag_output_bytes.decode('utf-8')

        if self.data_dict["tag_dict"][tag_name]["TYPE"]:
            if self.data_dict["tag_dict"][tag_name]["TYPE"] == "float":
                tag_output = float(tag_output)

            elif self.data_dict["tag_dict"][tag_name]["TYPE"] == "int":
                tag_output = int(tag_output)

            elif self.data_dict["tag_dict"][tag_name]["TYPE"] == "string":
                tag_output = str(tag_output)

        return tag_output

    def read_all_bytes(self):
        tag_start_pos = self.dict_end_pos
        tag_end_pos = self.data_dict["tag_len"] + self.dict_end_pos
        output_bytes = bytes(self.shm.buf[tag_start_pos:tag_end_pos])
        return output_bytes

    def write_to_tag(self, tag_name, tag_data):
        write_to_data_structure(self, tag_name, tag_data)

    def consume_write_all(self, consume_data):
        tag_start_pos = self.dict_end_pos
        tag_end_pos = self.data_dict["tag_len"] + self.dict_end_pos
        self.shm.buf[tag_start_pos:tag_end_pos] = consume_data

    def close(self):
        self.shm.close()

    def unlink(self):
        self.shm.unlink()



from comram import ramshare
import unittest
import xml.etree.ElementTree as ET


class Test_ramshare(unittest.TestCase):
    def test_init_read_write(self):
        test = ramshare.RamShare("test_share_name", "test_structure")
        test.write_to_tag("tag_1", 99)
        test_2 = int(test.read_tag("tag_1"))
        self.assertEqual(test_2, 99)
        test.unlink()

    def test_length_exception(self):
        test = ramshare.RamShare("test_share_name", "test_structure")
        self.assertRaises(ValueError, test.write_to_tag, "tag_1", 99999999999)
        test.unlink()

    def test_no_data_type(self):
        test = ramshare.RamShare("test_share_name", "test_structure")
        test.write_to_tag("tag_1", 99)
        test.close()
        test_2 = ramshare.RamShare("test_share_name")
        test_3 = int(test_2.read_tag("tag_1"))
        self.assertEqual(test_3, 99)
        test_2.unlink()

    def test_data_type_path(self):
        data_type_path = "data_structures/test_structure.xml"
        test = ramshare.RamShare("test_share_name_2", data_type_path)
        test.write_to_tag("tag_1", 99)
        test_2 = int(test.read_tag("tag_1"))
        self.assertEqual(test_2, 99)
        test.unlink()

    def test_close(self):
        test = ramshare.RamShare("test_share_name", "test_structure")
        test.write_to_tag("tag_1", 99)
        test.close()
        test.unlink()

    def test_data_type_file(self):
        data_structure_path = "data_structures/test_structure.xml"
        data_type = ET.parse(data_structure_path)
        test = ramshare.RamShare("test_share_name_2", data_type_file=data_type)
        test.write_to_tag("tag_1", 99)
        test_2 = int(test.read_tag("tag_1"))
        self.assertEqual(test_2, 99)
        test.unlink()

    def test_len_1_read_write(self):
        test = ramshare.RamShare("test_share_name", "test_structure")
        test.write_to_tag("tag_3", 3)
        test_2 = int(test.read_tag("tag_3"))
        self.assertEqual(test_2, 3)
        test.unlink()

if __name__ == '__main__':
    unittest.main()









import socket
import threading
import select
import time

from comram import ramshare


def producer_get_connection(self, producer_socket):
    timeout_in_seconds = 0.1
    connection_dict = {}
    while self.signal:
        ready = select.select([producer_socket], [], [], timeout_in_seconds)
        if ready[0]:
            msg, client_address = producer_socket.recvfrom(1024)

            consumer_name = msg[0:20].decode('utf-8').strip(" ")
            consumer_request = msg[20:30].decode('utf-8').strip(" ")

            if consumer_request == "connect":
                consumer_checksum = msg[30:62].decode('utf-8').strip(" ")
                checksum = self.share_memory.data_dict_checksum
                if consumer_checksum == checksum:
                    if self.debug:
                        print("checksum match")

                    if consumer_name in connection_dict:
                        old_client_address = connection_dict[consumer_name]
                        client_found = False
                        i = 0
                        while i < len(self.producer_client_list) and not client_found:
                            if self.producer_client_list[i] == old_client_address:
                                client_found = True
                                del self.producer_client_list[i]
                                print("deleted old address from consumer")
                            i += 1

                    self.producer_client_list.append(client_address)

                    respond_msg = "connection_success"
                    respond_msg = f"{respond_msg:<{20}}".encode('utf-8')
                    update_time = str(self.send_interval)
                    update_time = f"{update_time:<{10}}".encode('utf-8')
                    combine_msg = respond_msg + update_time
                    producer_socket.sendto(combine_msg, client_address)

                    connection_dict[consumer_name] = client_address
                    if self.debug:
                        consumer_name = msg.decode('utf-8')
                        print(f"{consumer_name} connection established")
                else:
                    respond_msg = "checksum_mismatch"
                    respond_msg = f"{respond_msg:<{20}}".encode('utf-8')
                    producer_socket.sendto(respond_msg, client_address)
                    print("checksum mismatch, do something more")

            if consumer_request == "disconnect":
                client_found = False
                i = 0
                while i < len(self.producer_client_list) and not client_found:
                    if self.producer_client_list[i] == client_address:
                        client_found = True
                        del self.producer_client_list[i]
                        print("disconnected consumer")
                    i += 1
                connection_dict.pop(consumer_name)


def producer_message_transmitter(self):
    producer_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    producer_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 65535)
    socket_address = (self.ip, self.port)
    producer_socket.bind(socket_address)

    t_transmitter = threading.Thread(target=producer_get_connection, args=(self, producer_socket,))
    t_transmitter.daemon = True
    t_transmitter.start()

    while self.signal:
        if len(self.producer_client_list):
            msg = self.share_memory.read_all_bytes()

            for i in range(len(self.producer_client_list)):
                producer_socket.sendto(msg, self.producer_client_list[i])
                if self.debug:
                    print("sending message to: ", self.producer_client_list[i])
        time.sleep(self.send_interval)


def consume_message_receiver(self):
    timeout_in_seconds = 0.5
    consumer_name = f"{self.name:<{20}}".encode('utf-8')
    con_dis = "connect"
    connect = f"{con_dis:<{10}}".encode('utf-8')

    checksum = self.share_memory.data_dict_checksum.encode('utf-8')
    message = consumer_name + connect + checksum
    while self.signal:
        consume_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        consume_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 65535)
        consume_socket.sendto(message, (self.ip, self.port))
        ready = select.select([consume_socket], [], [], timeout_in_seconds)
        if ready[0]:
            msg, _ = consume_socket.recvfrom(65535)
            connection_status = msg[0:20].decode('utf-8').strip(" ")
            update_time = float(msg[20:30].decode('utf-8').strip(" "))
            timeout_in_seconds = update_time * 100
            self.connection = True
            if connection_status == "connection_success":
                if self.debug:
                    print("successfully connected to producer")
            else:
                raise ValueError(connection_status)
        else:
            consume_socket.close()

        while self.signal and self.connection:
            ready = select.select([consume_socket], [], [], timeout_in_seconds)
            if ready[0]:
                msg, _ = consume_socket.recvfrom(65535)
                self.share_memory.consume_write_all(msg)
                if self.debug:
                    print(msg)
            else:
                self.share_memory.write_to_tag("connection_status", 2)
                if self.reconnect:
                    self.connection = False
                    consume_socket.close()
                else:
                    raise ConnectionError("Connection to producer timeout")

    if self.connection:
        con_dis = "disconnect"
        connect = f"{con_dis:<{10}}".encode('utf-8')
        disconnect_msg = consumer_name + connect
        consume_socket.sendto(disconnect_msg, (self.ip, self.port))
        consume_socket.close()


class Produce(threading.Thread):
    def __init__(self, share_name, ip=None, port=None, send_interval=0.01, data_type=None, debug=False):
        threading.Thread.__init__(self)
        self.signal = True
        self.ip = ip
        self.port = port
        self.data_type = data_type
        self.producer_client_list = []
        self.send_interval = send_interval
        self.share_name = share_name
        self.share_memory = ramshare.RamShare(self.share_name, self.data_type)
        self.debug = debug

    def start_produce(self):
        self.signal = True
        if self.port and self.ip and self.share_name:
            t1 = threading.Thread(target=producer_message_transmitter, args=(self,))
            t1.daemon = True
            t1.start()

        else:
            if not self.port:
                print("port required")
                raise ValueError

            if not self.ip:
                print("ip address required")
                raise ValueError

    def stop_produce(self):
        self.signal = False

    def unlink(self):
        self.share_memory.unlink()


class Consume(threading.Thread):

    def __init__(self, share_name, conname, ip=None, port=None, data_type=None, debug=False, reconnect=True):
        threading.Thread.__init__(self)
        self.signal = True
        self.ip = ip
        self.port = port
        self.data_type = data_type
        self.share_name = share_name
        self.share_memory = ramshare.RamShare(self.share_name, self.data_type)
        self.name = conname
        self.debug = debug
        self.connection = False
        self.reconnect = reconnect

    def start_consume(self):
        self.signal = True
        if self.port and self.ip:
            t1 = threading.Thread(target=consume_message_receiver, args=(self,))
            t1.daemon = True
            t1.start()

        else:
            if not self.port:
                raise ValueError("port required")

            if not self.ip:
                raise ValueError("ip address required")

            if not self.data_type:
                raise ValueError("data type required")

    def stop_consume(self):
        self.signal = False

    def unlink(self):
        self.share_memory.unlink()



from comram import ramshare
from comram import procon
import unittest
import time


class Test_procon(unittest.TestCase):
    def test_init_read_write(self):
        producer = procon.Produce("test_time_name2", data_type="time_structure", ip="127.0.0.1", port=9980,
                                  send_interval=0.01, debug=False)
        producer.start_produce()
        test_write = ramshare.RamShare("test_time_name2")
        test_write.write_to_tag("tag_3", 9999.9999)

        consumer = procon.Consume("test_time_name_3", "test_consumer", data_type="time_structure", ip="127.0.0.1",
                                  port=9980, debug=False)
        consumer.start_consume()
        test_read = ramshare.RamShare("test_time_name_3")

        time.sleep(0.1)
        tag_3_value = test_read.read_tag("tag_3")

        self.assertEqual(tag_3_value, 9999.9999)

        test_write.close()
        test_read.close()

        producer.stop_produce()
        producer.unlink()
        consumer.stop_consume()
        consumer.unlink()

    def test_consume_reconnect(self):
        consumer = procon.Consume("test_time_name_4", "test_consumer", data_type="time_structure", ip="127.0.0.1",
                                  port=9981)
        consumer.start_consume()
        test_read = ramshare.RamShare("test_time_name_4")

        time.sleep(3)

        producer = procon.Produce("test_time_name5", data_type="time_structure", ip="127.0.0.1", port=9981,
                                  send_interval=0.01)
        producer.start_produce()

        test_write = ramshare.RamShare("test_time_name5")
        test_write.write_to_tag("tag_3", 9999.9999)

        time.sleep(2)
        tag_3_value = test_read.read_tag("tag_3")

        self.assertEqual(tag_3_value, 9999.9999)

        test_write.close()
        test_read.close()

        producer.stop_produce()
        producer.unlink()
        consumer.stop_consume()
        consumer.unlink()

    def test_connection_timeout(self):
        producer = procon.Produce("test_time_name2", data_type="time_structure", ip="127.0.0.1", port=9980,
                                  send_interval=0.01, debug=False)
        producer.start_produce()
        test_write = ramshare.RamShare("test_time_name2")
        test_write.write_to_tag("tag_3", 9999.9999)

        consumer = procon.Consume("test_time_name_3", "test_consumer", data_type="time_structure", ip="127.0.0.1",
                                  port=9980, debug=False)
        consumer.start_consume()
        test_read = ramshare.RamShare("test_time_name_3")

        time.sleep(0.1)
        tag_3_value = test_read.read_tag("tag_3")

        self.assertEqual(tag_3_value, 9999.9999)

        test_write.close()
        producer.stop_produce()
        producer.unlink()

        time.sleep(2)

        connection_status = test_read.read_tag("connection_status")
        connection_status = connection_status.strip(" ")
        self.assertEqual(connection_status, "connection_timeout")

        test_read.close()
        consumer.stop_consume()
        consumer.unlink()


if __name__ == '__main__':
    unittest.main()
